

from brownie import chain
from dank_mids import instances
from multicall import Call
from multicall.utils import await_awaitable, gather

from tests.fixtures import dank_w3

CHAI = '0x06AF07097C9Eeb7fD685c692751D5C66dB49c215'
height = chain.height
BIG_WORK = [Call(CHAI, 'totalSupply()(uint)', [[f'totalSupply{i}',None]], block_id=height - (i // 25000), _w3=dank_w3).coroutine() for i in range(100_000)]
height = chain.height
MULTIBLOCK_WORK = [Call(CHAI, 'totalSupply()(uint)', [[f'totalSupply{i}',None]], _w3=dank_w3, block_id=height-i).coroutine() for i in range(100_000)]

def test_dank_middleware():
    await_awaitable(gather(BIG_WORK))
    dank_mids_controller = instances[0]
    cid = dank_mids_controller.call_uid.latest
    mid = dank_mids_controller.worker.multicall_uid.latest
    rid = dank_mids_controller.worker.request_uid.latest
    assert cid, "The DankMiddlewareController did not process any calls."
    assert mid, "The DankMiddlewareController did not process any batches."
    assert rid, "The DankMiddlewareController did not process any requests."
    print(f"calls:                  {cid}")
    print(f"multicalls:             {mid}")
    print(f"requests:               {rid}")
    print(f"calls per multicall:    {cid/mid}")
    print(f"calls per request:      {cid/rid}")
    print(f"multicalls per request: {mid/rid}")
    # General "tests" that verify batching performance
    assert mid < cid / 100, f"Batched {cid} calls into {mid} multicalls. Performance underwhelming."
    assert rid < cid / 150,  f"Batched {cid} calls into {rid} requests. Performance underwhelming."
    assert mid / rid > 1,  f"Batched {mid} multicalls into {rid} requests. Performance underwhelming."

def test_bad_hex_handling():
    chainlinkfeed = "0xfe67209f6FE3BA6cE36d0941700085C194e958DF"
    assert await_awaitable(Call(chainlinkfeed, 'latestAnswer()(uint)', block_id=14_000_000).coroutine()) == 15717100
    assert chainlinkfeed in instances[0].do_not_batch

def test_next_cid():
    assert instances[0].call_uid.next + 1 == instances[0].call_uid.next
    
def test_next_mid():
    assert instances[0].worker.request_uid.next + 1 == instances[0].worker.request_uid.next
    
def test_next_bid():
    assert instances[0].worker.multicall_uid.next + 1 == instances[0].worker.multicall_uid.next



from typing import TYPE_CHECKING, Any, Callable, Coroutine, Dict, List, Literal, NewType, TypedDict, Union
from eth_typing import ChecksumAddress

from web3.types import RPCEndpoint, RPCResponse

if TYPE_CHECKING:
    from dank_mids.call import BatchedCall

BlockId = NewType("BlockId", str)
CallsToExec = Dict[BlockId, List["BatchedCall"]]
MulticallBatch = List["BatchedCall"]
JsonrpcBatch = List[MulticallBatch]

eth_callParams = TypedDict("eth_callParams", {"to": ChecksumAddress, "data": str})
OverrideParams = TypedDict("OverrideParams", {"code": str})
JsonrpcParams = List[Union[eth_callParams, BlockId, OverrideParams]]

JsonrpcCall = TypedDict(
    "JsonrpcCall", 
    {
        "jsonrpc": str,
        "id": int,
        "method": Literal["eth_call"],
        "params": JsonrpcParams,
    }
)

# This type alias was introduced in web3 v5.28.0 but we like loose deps here so we recreate instead of import.
AsyncMiddleware = Callable[[RPCEndpoint, Any], Coroutine[Any, Any, RPCResponse]]




import os

from aiohttp import ClientTimeout

LOOP_INTERVAL = float(os.environ.get("DANKMIDSLOOPINTERVAL", 0.01))

# Enable Demo Mode?
DEMO_MODE = bool(os.environ.get("DANKMIDS_DEMO_MODE")) or bool(os.environ.get("DANK_MIDS_DEMO_MODE"))

# With default AsyncBaseProvider settings, some dense calls will fail
#  due to aiohttp.TimeoutError where they would otherwise succeed.
# We set the default to 2 minutes but if you're doing serious work
#  you may want to increase it further.
AIOHTTP_TIMEOUT = ClientTimeout(int(os.environ.get("AIOHTTP_TIMEOUT", 120)))
